import json
import logging
import os

from shutil import rmtree

from nearuplib.constants import NODE_PID_FILE, LOCALNET_LOGS_FOLDER
from nearuplib.nodelib import run_binary, proc_name_from_pid, check_exist_neard


def run(binary_path, home, num_nodes, num_shards, override, verbose=True):
    if override:
        if os.path.exists(home):
            logging.info("Removing old data.")
            rmtree(home)

    if not os.path.exists(home):
        run_binary(binary_path,
                   home,
                   'testnet',
                   shards=num_shards,
                   validators=num_nodes).wait()

    # Edit args files
    for i in range(0, num_nodes):
        args_json = os.path.join(home, f'node{i}', 'config.json')

        with open(args_json, 'r') as config_file:
            data = json.load(config_file)
        data['rpc']['addr'] = f'0.0.0.0:{3030 + i}'
        data['network']['addr'] = f'0.0.0.0:{24567 + i}'

        with open(args_json, 'w') as config_file:
            json.dump(data, config_file, indent=2)

    # Load public key from first node
    with open(os.path.join(home, 'node0', 'node_key.json'),
              'r') as node_key_file:
        data = json.load(node_key_file)
        public_key = data['public_key']

    # Recreate log folder
    rmtree(LOCALNET_LOGS_FOLDER, ignore_errors=True)
    os.mkdir(LOCALNET_LOGS_FOLDER)

    # Spawn network
    pid_fd = open(NODE_PID_FILE, 'w')
    for i in range(0, num_nodes):
        proc = run_binary(
            binary_path,
            os.path.join(home, f'node{i}'),
            'run',
            verbose=verbose,
            boot_nodes=f'{public_key}@127.0.0.1:24567' if i > 0 else None,
            output=os.path.join(LOCALNET_LOGS_FOLDER, f'node{i}'))
        proc_name = proc_name_from_pid(proc.pid)
        pid_fd.write(f"{proc.pid}|{proc_name}|localnet\n")
    pid_fd.close()

    logging.info("Localnet was spawned successfully...")
    logging.info(f"Localnet logs written in: {LOCALNET_LOGS_FOLDER}")
    logging.info("Check localnet status at http://127.0.0.1:3030/status")


def entry(binary_path, home, num_nodes, num_shards, override, verbose):
    if binary_path:
        binary_path = os.path.join(binary_path, 'neard')

    check_exist_neard()

    run(binary_path, home, num_nodes, num_shards, override, verbose)



import os
import shutil
import time

import pytest
import requests

from nearuplib.nodelib import restart_nearup, setup_and_run, stop_nearup
from nearuplib.constants import LOGS_FOLDER

NEAR_DIR = '~/.near'
NEARUP_DIR = '~/.nearup'

NEARUP_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    '../nearup',
)


def cleanup():
    if os.path.exists(NEAR_DIR):
        shutil.rmtree(NEAR_DIR)

    if os.path.exists(NEARUP_DIR):
        shutil.rmtree(NEARUP_DIR)

    if os.path.exists(LOGS_FOLDER):
        shutil.rmtree(LOGS_FOLDER)


def setup_module(module):  # pylint: disable=W0613
    cleanup()

    if not os.path.exists(LOGS_FOLDER):
        os.makedirs(LOGS_FOLDER)


def teardown_module(module):  # pylint: disable=W0613
    cleanup()


def test_nearup_still_runnable():
    setup_and_run(binary_path='',
                  home_dir='',
                  init_flags=['--chain-id=betanet'],
                  boot_nodes='',
                  verbose=True,
                  no_watcher=True)
    time.sleep(10)

    resp = requests.get('http://localhost:3030/status')
    assert resp.status_code == 200
    assert resp.text

    stop_nearup()
    time.sleep(10)

    with pytest.raises(requests.exceptions.ConnectionError):
        requests.get('http://localhost:3030/status')

    restart_nearup('betanet', NEARUP_PATH, watcher=False)
    time.sleep(10)

    resp = requests.get('http://localhost:3030/status')
    assert resp.status_code == 200
    assert resp.text

    stop_nearup()
    time.sleep(10)

    with pytest.raises(requests.exceptions.ConnectionError):
        requests.get('http://localhost:3030/status')



import hashlib
import json
import logging
import os
import shutil
import subprocess
import sys

from subprocess import Popen

import psutil

from nearuplib.constants import LOGS_FOLDER, NODE_PID_FILE
from nearuplib.util import (
    download_binaries,
    download_config,
    download_genesis,
    initialize_keys,
    latest_genesis_md5sum,
)
from nearuplib.watcher import run_watcher, stop_watcher


def init(home_dir, binary_path, init_flags):
    logging.info("Initializing the node configuration using near binary...")
    target = f'{binary_path}/near'
    subprocess.call([target, '--home=%s' % home_dir, 'init'] + init_flags)


def init_official(chain_id, binary_path, home_dir, account_id):
    logging.info("Initializing the keys...")
    initialize_keys(home_dir, binary_path, account_id)

    logging.info("Downloading the genesis file...")
    download_genesis(chain_id, home_dir)

    logging.info("Downloading the config file...")
    download_config(chain_id, home_dir)


def get_chain_id_from_flags(flags):
    """Retrieve requested chain id from the flags."""
    chain_id_flags = [flag for flag in flags if flag.startswith('--chain-id=')]
    if len(chain_id_flags) == 1:
        return chain_id_flags[0][len('--chain-id='):]
    return ''


def genesis_changed(chain_id, home_dir):
    genesis_md5sum = latest_genesis_md5sum(chain_id)
    local_genesis_md5sum = hashlib.md5(
        open(os.path.join(os.path.join(home_dir, 'genesis.json')),
             'rb').read()).hexdigest()
    if genesis_md5sum == local_genesis_md5sum:
        logging.info('The genesis version is up to date')
        return False

    logging.info(
        f'Remote genesis protocol version md5 {genesis_md5sum}, ours is {local_genesis_md5sum}'
    )
    return True


def check_and_update_genesis(chain_id, home_dir):
    if genesis_changed(chain_id, home_dir):
        logging.info(
            f'Update genesis config and remove stale data for {chain_id}')
        os.remove(os.path.join(home_dir, 'genesis.json'))
        download_genesis(chain_id, home_dir)
        if os.path.exists(os.path.join(home_dir, 'data')):
            shutil.rmtree(os.path.join(home_dir, 'data'))
        return True
    return False


def check_and_setup(binary_path, home_dir, init_flags):
    """Checks if there is already everything setup on this machine, otherwise sets up NEAR node."""
    chain_id = get_chain_id_from_flags(init_flags)
    if os.path.exists(os.path.join(home_dir)):
        missing = []
        for file in ['node_key.json', 'config.json', 'genesis.json']:
            if not os.path.exists(os.path.join(home_dir, file)):
                missing.append(file)
        if missing:
            logging.error(
                f'Missing files {", ".join(missing)} in {home_dir}. Maybe last init was failed.'
            )
            logging.error(
                'Either specify different --home or remove {home_dir} to start from scratch.'
            )
            sys.exit(1)

        genesis_config = json.loads(
            open(os.path.join(os.path.join(home_dir, 'genesis.json'))).read())
        if genesis_config['chain_id'] != chain_id:
            logging.error(
                f"Folder {home_dir} has network configuration for {genesis_config['chain_id']}"
                f"Either specify different --home or remove {home_dir} to start from scratch.",
            )
            sys.exit(1)

        if chain_id in ['betanet', 'testnet']:
            check_and_update_genesis(chain_id, home_dir)
        else:
            logging.info("Using existing node configuration from %s for %s",
                         home_dir, genesis_config['chain_id'])
        return

    logging.info("Setting up network configuration.")
    account_id = [x for x in init_flags if x.startswith('--account-id')]
    if not account_id:
        logging.warning(
            "To be a validator please specify the --account-id flag with the validator id."
        )
        account_id = ""
    else:
        account_id = account_id[0].split('=')[-1]

    if chain_id in ['betanet', 'testnet']:
        init_official(chain_id, binary_path, home_dir, account_id)
    else:
        init(home_dir, binary_path, init_flags)

    if chain_id not in ['betanet', 'testnet']:
        filename = os.path.join(home_dir, 'genesis.json')
        genesis_config = json.load(open(filename))
        genesis_config['gas_price'] = 0
        genesis_config['min_gas_price'] = 0
        json.dump(genesis_config, open(filename, 'w'))


def print_staking_key(home_dir):
    key_path = os.path.join(home_dir, 'validator_key.json')
    if not os.path.exists(key_path):
        return

    key_file = json.loads(open(key_path).read())
    if not key_file['account_id']:
        logging.warning(
            "Node is not staking. Re-run init to specify staking account.")
        return
    logging.info("Stake for user '%s' with '%s'", key_file['account_id'],
                 key_file['public_key'])


def run_binary(path,
               home,
               action,
               verbose=None,
               shards=None,
               validators=None,
               non_validators=None,
               boot_nodes=None,
               output=None):
    command = [path, '--home', home]

    if verbose or verbose == '':
        command.extend(['--verbose', verbose])

    command.append(action)

    if shards:
        command.extend(['--shards', str(shards)])
    if validators:
        command.extend(['--v', str(validators)])
    if non_validators:
        command.extend(['--n', str(non_validators)])
    if boot_nodes:
        command.extend(['--boot-nodes', boot_nodes])

    if output:
        output = open(f'{output}.log', 'a')

    near = Popen(command, stderr=output, stdout=output)
    return near


def proc_name_from_pid(pid):
    process = psutil.Process(pid)
    return process.name()


def check_exist_neard():
    if os.path.exists(NODE_PID_FILE):
        logging.warning(
            "There is already binary nodes running. Stop it using: nearup stop")
        logging.warning(f"If this is a mistake, remove {NODE_PID_FILE}")
        sys.exit(1)


def run(home_dir, binary_path, boot_nodes, verbose, chain_id, watch=False):
    os.environ['RUST_BACKTRACE'] = '1'
    # convert verbose = True to --verbose '' command line argument
    if verbose:
        verbose = ''
    else:
        verbose = None
    proc = run_binary(os.path.join(binary_path, 'near'),
                      home_dir,
                      'run',
                      verbose=verbose,
                      boot_nodes=boot_nodes,
                      output=os.path.join(LOGS_FOLDER, chain_id))
    proc_name = proc_name_from_pid(proc.pid)

    with open(NODE_PID_FILE, 'w') as pid_fd:
        pid_fd.write(f"{proc.pid}|{proc_name}|{chain_id}")
        pid_fd.close()

    logging.info("Node is running...")
    logging.info("To check logs call: `nearup logs` or `nearup logs --follow`")

    if watch:
        logging.info("Watcher is enabled. Starting watcher...")
        run_watcher(chain_id)


def setup_and_run(binary_path,
                  home_dir,
                  init_flags,
                  boot_nodes,
                  verbose=False,
                  no_watcher=False):
    check_exist_neard()
    chain_id = get_chain_id_from_flags(init_flags)

    if binary_path == '':
        logging.info('Using officially compiled binary')
        uname = os.uname()[0]
        if uname not in ['Linux']:
            logging.error(
                'Sorry your Operating System does not have officially compiled binary now.'
            )
            logging.error(
                'Compile a local binary and set --binary-path when running')
            sys.exit(1)
        binary_path = os.path.expanduser(f'~/.nearup/near/{chain_id}')

        if not os.path.exists(binary_path):
            os.makedirs(binary_path)

        download_binaries(chain_id, uname)
        watch = True
    else:
        logging.info(f'Using local binary at {binary_path}')

    check_and_setup(binary_path, home_dir, init_flags)

    print_staking_key(home_dir)

    if no_watcher:
        watch = False

    run(home_dir, binary_path, boot_nodes, verbose, chain_id, watch=watch)


def stop_nearup(keep_watcher=False):
    logging.warning("Stopping the near daemon...")
    stop_native()

    if not keep_watcher:
        logging.warning("Stopping the nearup watcher...")
        stop_watcher()
    else:
        logging.warning("Skipping the stopping of the nearup watcher...")


def restart_nearup(net,
                   path=os.path.expanduser('~/.local/bin/nearup'),
                   watcher=True):
    logging.warning("Restarting nearup...")

    if not os.path.exists(path):
        logging.error(
            "Please delete current nearup and install the new with `pip3 install --user nearup`"
        )
        logging.error(
            "For local development use: `pip3 install --user .` from root directory"
        )
        sys.exit(1)

    logging.warning("Stopping nearup...")
    stop_nearup(keep_watcher=watcher)

    logging.warning("Starting nearup...")
    setup_and_run(binary_path='',
                  home_dir='',
                  init_flags=[f'--chain-id={net}'],
                  boot_nodes='',
                  verbose=True,
                  no_watcher=not watcher)

    logging.info("Nearup has been restarted...")


def stop_native():
    try:
        if os.path.exists(NODE_PID_FILE):
            with open(NODE_PID_FILE) as pid_file:
                for line in pid_file.readlines():
                    pid, proc_name, _ = line.strip().split("|")
                    pid = int(pid)
                    process = psutil.Process(pid)
                    logging.info(
                        f"Near procces is {proc_name} with pid: {pid}...")
                    if proc_name in proc_name_from_pid(pid):
                        logging.info(
                            f"Stopping process {proc_name} with pid {pid}...")
                        process.kill()
            os.remove(NODE_PID_FILE)
        else:
            logging.info("Near deamon is not running...")
    except Exception as ex:
        logging.error(f"There was an error while stopping watcher: {ex}")
        if os.path.exists(NODE_PID_FILE):
            os.remove(NODE_PID_FILE)



import json
import os
import shutil

from nearuplib.util import download_binaries, download_config, download_genesis

HOME_DIR = os.path.expanduser('~/.near/betanet')
NEARUP_BINARY_DIR = os.path.expanduser('~/.nearup/near/betanet')


def setup_module(module):  # pylint: disable=W0613
    if os.path.exists(HOME_DIR):
        shutil.rmtree(HOME_DIR)

    if os.path.exists(NEARUP_BINARY_DIR):
        shutil.rmtree(NEARUP_BINARY_DIR)

    os.makedirs(HOME_DIR)
    os.makedirs(NEARUP_BINARY_DIR)


def test_download_binaries():
    download_binaries('betanet', 'Linux')

    expected_binaries = ['near', 'keypair-generator', 'genesis-csv-to-json']
    for binary in expected_binaries:
        path = os.path.join(NEARUP_BINARY_DIR, binary)
        # check if the binary exists
        assert os.path.exists(path)
        # check if the binary is executable
        assert os.access(path, os.X_OK)


def test_download_config():
    download_config('betanet', HOME_DIR)
    path = os.path.join(HOME_DIR, 'config.json')

    assert os.path.exists(path)
    with open(path) as config:
        json.loads(config.read())


def test_download_genesis():
    download_genesis('betanet', HOME_DIR)
    path = os.path.join(HOME_DIR, 'genesis.json')

    assert os.path.exists(path)
    with open(path) as genesis:
        json.loads(genesis.read())



import logging
import os
import sys

from subprocess import Popen

import psutil

from nearuplib.constants import WATCHER_PID_FILE


def run_watcher(net, path=os.path.expanduser('~/.local/bin/watcher')):
    logging.info("Starting the nearup watcher...")

    if not os.path.exists(path):
        logging.error(
            "Delete current nearup and install the new with `pip3 install --user nearup`"
        )
        logging.error(
            "To run nearup locally use: `pip3 install --user .` from root directory"
        )
        sys.exit(1)

    proc = Popen(['python3', path, 'run', net])

    with open(WATCHER_PID_FILE, 'w') as watcher_pid_file:
        watcher_pid_file.write(str(proc.pid))


def stop_watcher():
    try:
        if os.path.exists(WATCHER_PID_FILE):
            with open(WATCHER_PID_FILE) as pid_file:
                pid = int(pid_file.read())
                process = psutil.Process(pid)
                logging.info(
                    f'Stopping near watcher {process.name()} with pid {pid}...')
                process.kill()
                os.remove(WATCHER_PID_FILE)
        else:
            logging.info("Nearup watcher is not running...")
    except Exception as ex:
        logging.error(f"There was an error while stopping watcher: {ex}")



import logging
import os
import subprocess
import sys

from nearuplib.constants import NODE_PID_FILE


def show_logs(follow, number_lines):
    if not os.path.exists(NODE_PID_FILE):
        logging.info('Node is not running')
        sys.exit(1)

    pid_info = open(NODE_PID_FILE).readline()
    logging.info(pid_info)
    _, _, network = pid_info.strip().split("|")

    if not network in ['betanet', 'testnet']:
        # TODO: localnet could have several logs, not showing them all but list log files here
        # Maybe better to support `nearup logs node0` usage.
        logging.info(
            'You are running local net. Logs are in: ~/.nearup/localnet-logs/')
        sys.exit(0)

    command = [
        'tail',
        '-n',
        str(number_lines),
        '-f' if follow else '',
        os.path.expanduser(f'~/.nearup/logs/{network}.log'),
    ]

    try:
        subprocess.run(command, start_new_session=True, check=True)
    except KeyboardInterrupt:
        sys.exit(0)
    except subprocess.CalledProcessError:
        logging.error("Unable to read logs. Please try again.")
        sys.exit(1)



import logging
import os
import stat
import subprocess
import sys

import boto3
from botocore import UNSIGNED
from botocore.client import Config

from nearuplib.constants import S3_BUCKET


def download_from_s3(path, filepath=None):
    s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
    s3_client.download_file(S3_BUCKET, path, filepath)


def read_from_s3(path):
    s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
    response = s3_client.get_object(Bucket=S3_BUCKET, Key=path)
    return response['Body'].read().decode('utf-8')


def download_config(net, home_dir):
    download_from_s3(f'nearcore-deploy/{net}/config.json',
                     os.path.join(home_dir, 'config.json'))


def download_genesis(net, home_dir):
    download_from_s3(f'nearcore-deploy/{net}/genesis.json',
                     os.path.join(home_dir, 'genesis.json'))


def download_binaries(net, uname):
    commit = latest_deployed_release_commit(net)
    branch = latest_deployed_release_branch(net)

    if commit:
        logging.info(f'Downloading latest deployed version for {net}')

        binaries = ['near', 'keypair-generator', 'genesis-csv-to-json']
        for binary in binaries:
            download_url = f'nearcore/{uname}/{branch}/{commit}/{binary}'
            download_path = os.path.expanduser(f'~/.nearup/near/{net}/{binary}')

            logging.info(
                f"Downloading {binary} to {download_path} from {download_url}..."
            )
            download_from_s3(download_url, download_path)
            logging.info(f"Downloaded {binary} to {download_path}...")

            logging.info(f"Making the {binary} executable...")
            status = os.stat(download_path)
            os.chmod(download_path, status.st_mode | stat.S_IEXEC)

        # TODO: seperate into download_metadata function with missing metadata
        with open(os.path.expanduser(f'~/.nearup/near/{net}/version'),
                  'w') as version_file:
            version_file.write(commit)


def latest_deployed_release_commit(net):
    return read_from_s3(f'nearcore-deploy/{net}/latest_deploy').strip()


def latest_deployed_release_commit_has_changed(net, commit):
    latest_commit = latest_deployed_release_commit(net)

    logging.info(f"Current release commit is: {commit}")
    logging.info(f"Latest release commit is {latest_commit}")

    if not commit:
        return False

    return commit != latest_commit


def latest_deployed_release_branch(net):
    return read_from_s3(f'nearcore-deploy/{net}/latest_release').strip()


def latest_deployed_release_time(net):
    return read_from_s3(f'nearcore-deploy/{net}/latest_deploy_at').strip()


def latest_genesis_md5sum(net):
    return read_from_s3(f'nearcore-deploy/{net}/genesis_md5sum').strip()


def generate_key(cmd, key):
    logging.info(f"Generating {key}...")

    try:
        subprocess.check_call(cmd, stdout=subprocess.PIPE)
    except KeyboardInterrupt:
        logging.error("\nStopping NEARCore.")
    except (subprocess.CalledProcessError, FileNotFoundError):
        logging.error(f"Unable to generate {key}.")
        sys.exit(1)

    logging.info(f"{key.capitalize()} generated...")


def generate_node_key(home, binary_path):
    cmd = [
        f'{binary_path}/keypair-generator', '--home', home, '--generate-config',
        'node-key'
    ]
    generate_key(cmd, 'node key')


def generate_validator_key(home, binary_path, account_id):
    cmd = [
        f'{binary_path}/keypair-generator', '--home', home, '--generate-config',
        '--account-id', account_id, 'validator-key'
    ]
    generate_key(cmd, 'validator key')


def initialize_keys(home, binary_path, account_id):
    logging.info("Generating the node keys...")
    generate_node_key(home, binary_path)

    if account_id:
        logging.info("Generating the validator keys...")
        generate_validator_key(home, binary_path, account_id)
